COMMAND_MAP = {
                  'dollar': '$',
                  'braceright': '}',
                  'braceleft': '{',
                  'bracketright': ']',
                  'bracketleft': '[',
                  'parenright': ')',
                  'parenleft': '(',
                  'colon': ':',
                  'semicolon': ';',
                  'bar': '|',
                  'greater': '>',
                  'less': '<',
                  'comma': ',',
                  'period': '.',
                  'slash': '/',
                  'question': '?',
                  'plus': '+',
                  'equal': '=',
                  'minus': '-',
                  'underscore': '_',
                  'exclam': '!',
                  'at': '@',
                  'percent': '%',
                  'asciicircum': '^',
                  'ampersand': '&',
                  'asterisk': '*',
                  'quoteright': "'",
                  'quotedbl': '"',
                  'BackSpace': 'BackSpace',
                  'Return': 'Return',
                  'space': ' ',
                  'Up': '<Up>',
                  'Down': '<Down>',
                  'Left': '<Left>',
                  'Right': '<Right>',
                  '<Control-braceright>': '<Control-bracketright>',
                  '<Control-braceleft>': '<Control-bracketleft>',
               }


DEFAULT_MOVEMENTS = {
                        'p': ['paste'],
                        'u': ['undo_command'],
                        'k': ['move_cursor_up'],
                        'h': ['move_cursor_left'],
                        'j': ['move_cursor_down'],
                        'l': ['move_cursor_right'],
                        'A': ['insert_end_of_line'],
                        'G': ['move_cursor_end_file'],
                        '$': ['move_cursor_end_line'],
                        'O': ['insert_new_line_above'],
                        'o': ['insert_new_line_below'],
                        'x': ['delete_text_highlight'],
                        '<Control-r>': ['redo_command'],
                        'gg': ['move_cursor_begin_file'],
                        'e': ['move_cursor_next_word_end'],
                        '}': ['move_cursor_next_paragraph'],
                        '{': ['move_cursor_prev_paragraph'],
                        '0': ['move_cursor_beginning_line'],
                        'w': ['move_cursor_next_word_front'],
                        'b': ['move_cursor_prev_word_front'],
                        '<Control-braceright>': ['move_next_instance_buffer'],
                        '<Control-braceleft>': ['move_prev_instance_buffer'],
                    }

DEFAULT_COMMAND_LEADERS = set(['g', 'f', 'd', 'y'])

VISUAL_MOVEMENTS = {
                        'k': ['move_cursor_up'],
                        'h': ['move_cursor_left'],
                        'j': ['move_cursor_down'],
                        'l': ['move_cursor_right'],
                        '<': ['shift_selection_left'],
                        'G': ['move_cursor_end_file'],
                        '$': ['move_cursor_end_line'],
                        '>': ['shift_selection_right'],
                        'x': ['delete_text_highlight'],
                        'gg': ['move_cursor_begin_file'],
                        'e': ['move_cursor_next_word_end'],
                        '}': ['move_cursor_next_paragraph'],
                        '{': ['move_cursor_prev_paragraph'],
                        '0': ['move_cursor_beginning_line'],
                        'w': ['move_cursor_next_word_front'],
                        'b': ['move_cursor_prev_word_front']
                    }




BREAK_MOVEMENTS = {
                      'h': ['move_cursor_left'],
                      'j': ['move_cursor_down'],
                      'k': ['move_cursor_up'],
                      'l': ['move_cursor_right'],
                      'w': ['move_cursor_next_word_front'],
                      'b': ['move_cursor_prev_word_front'],
                      'x': ['delete_text_highlight']
                  }



import interaction_manager, command_list, command_parser
from copy import deepcopy
from State import instance
import re

DEFAULT_MOVEMENTS = command_list.DEFAULT_MOVEMENTS
DEFAULT_COMMAND_LEADERS = command_list.DEFAULT_COMMAND_LEADERS
VISUAL_MOVEMENTS = command_list.VISUAL_MOVEMENTS
BREAK_MOVEMENTS = command_list.BREAK_MOVEMENTS
COMMAND_MAP = command_list.COMMAND_MAP

# parses keyboard input for meaningful instructions
# sends "completed" instructions to interaction_manager router
class user_input():
    def __init__(self):
        self.graphics = None
        self.curr_state, self.command_buffer = 'Default', ''
        self.instances, self.copy_buffer, self.curr_instance = [], [], 0

    def start_instance(self, filename):
        self.instances.append(instance.instance(filename))

    def set_GUI_reference(self, canvas):
        self.graphics = canvas
        self.instances[self.curr_instance].set_line_height(self.graphics.line_height)
        interaction_manager.render_page([], [], self.graphics, self.get_curr_instance(), self)

    def add_copy_buffer(self, l):
        self.copy_buffer = l

    def get_curr_instance(self):
        return self.instances[self.curr_instance]

    def get_copy_buffer(self):
        return self.copy_buffer

    def go_next_instance(self):
        if self.curr_instance < len(self.instances) - 1:
            self.curr_instance += 1
            self.set_GUI_reference(self.graphics)

    def go_prev_instance(self):
        if self.curr_instance > 0:
            self.curr_instance -= 1
            self.set_GUI_reference(self.graphics)

    # checks if key input an integer greater than 0 and less than 10
    def is_digit(self, k):
        return (len(k) == 1) and (ord(k) >= 49 and ord(k) <= 57)

    def key(self, event):
        key = event.keysym
        if key != '??':
            # if key is not in [a-zA-Z0-9] length of keysym will be greater than one
            if len(key) > 1:
                try:
                    k = COMMAND_MAP[key]
                    self.user_key_pressed(k)
                except KeyError:
                    pass
            else:
                self.user_key_pressed(key)

    def control_a(self, event):
        self.user_key_pressed('<Control-a>')

    def control_b(self, event):
        self.user_key_pressed('<Control-b>')

    def control_c(self, event):
        self.user_key_pressed('<Control-c>')

    def control_d(self, event):
        self.user_key_pressed('<Control-d>')

    def control_e(self, event):
        self.user_key_pressed('<Control-e>')

    def control_f(self, event):
        self.user_key_pressed('<Control-f>')

    def control_g(self, event):
        self.user_key_pressed('<Control-g>')

    def control_h(self, event):
        self.user_key_pressed('<Control-h>')

    def control_i(self, event):
        self.user_key_pressed('<Control-i>')

    def control_j(self, event):
        self.user_key_pressed('<Control-j>')

    def control_k(self, event):
        self.user_key_pressed('<Control-k>')

    def control_l(self, event):
        self.user_key_pressed('<Control-l>')

    def control_m(self, event):
        self.user_key_pressed('<Control-m>')

    def control_n(self, event):
        self.user_key_pressed('<Control-n>')

    def control_o(self, event):
        self.user_key_pressed('<Control-o>')

    def control_p(self, event):
        self.user_key_pressed('<Control-p>')
    def control_q(self, event):
        self.user_key_pressed('<Control-q>')

    def control_r(self, event):
        self.user_key_pressed('<Control-r>')

    def control_s(self, event):
        self.user_key_pressed('<Control-s>')

    def control_t(self, event):
        self.user_key_pressed('<Control-t>')

    def control_u(self, event):
        self.user_key_pressed('<Control-u>')

    def control_v(self, event):
        self.user_key_pressed('<Control-v>')

    def control_w(self, event):
        self.user_key_pressed('<Control-w>')

    def control_x(self, event):
        self.user_key_pressed('<Control-x>')

    def control_y(self, event):
        self.user_key_pressed('<Control-y>')

    def control_z(self, event):
        self.user_key_pressed('<Control-z>')

    def control_braceright(self, event):
        self.user_key_pressed('<Control-braceright>')

    def control_braceleft(self, event):
        self.user_key_pressed('<Control-braceleft>')

    def escape(self, event):
        self.curr_state = 'Default'
        self.command_buffer = ''
        interaction_manager.render_page([], [], self.graphics, self.instances[self.curr_instance], self)

    # TODO: THIS LOOKS HACKY
    def mouse_scroll(self, event):
        # run up or down command depending on scroll direction
        delta = event.delta * -1
        self.curr_state = 'Default'
        self.command_buffer = ''
        cmd = ['n' + str(delta), 'mouse_scroll']
        interaction_manager.input_command(cmd, self.graphics, self.get_curr_instance(), None)

    def user_key_pressed(self, key):
        if self.curr_state == 'Default':
            self.user_key_default(key)
        elif self.curr_state == 'Insert':
            self.user_key_insert(key)
        elif self.curr_state == 'Visual':
            self.user_key_visual(key)
        elif self.curr_state == 'Ex':
            self.user_key_ex(key)
        elif self.curr_state == 'fuzzy_file_selection':
            self.user_key_fuzzy_file_select(key)

    def init_insert_mode(self):
        self.curr_state = 'Insert'

    def init_ex_mode(self):
        self.curr_state = 'Ex'

    def init_visual_mode(self):
        curr_instance = self.get_curr_instance()
        # set once and then never mutate this ever again per visual selection
        self.get_curr_instance().set_visual_anchor()
        self.curr_state = 'Visual'

    def user_key_default(self, key):
        # BEGIN MODE CHANGE MAPPINGS HERE
        mode_dict = { 'i': self.init_insert_mode, 'v': self.init_visual_mode, ':': self.init_ex_mode, '<Control-p>': self.init_fuzzy_matching }
        # END MODE CHANGE MAPPINGS HERE
        # Command to be buffered
        if key in DEFAULT_COMMAND_LEADERS or self.is_digit(key) or len(self.command_buffer):
            self.command_buffer += key
            s_par = command_parser.default_parse(self.command_buffer)

            if s_par != '' or key in BREAK_MOVEMENTS:
                cmd = s_par if s_par != '' else BREAK_MOVEMENTS[key]
                interaction_manager.input_command(cmd, self.graphics, self.get_curr_instance(), self)
                self.command_buffer = ''

        # default movement requested
        elif key in DEFAULT_MOVEMENTS:
            interaction_manager.input_command(DEFAULT_MOVEMENTS[key], self.graphics, self.get_curr_instance(), self)
            self.command_buffer = ''
        # mode change requested
        elif key in mode_dict:
            mode_dict[key]()

    # this should be the only state that doesn't change no matter the configuration
    def user_key_insert(self, key):
        if not key in ['BackSpace', 'Return']:
            cmd = ['s' + key, 'insert_text']
            interaction_manager.input_command(cmd, self.graphics, self.get_curr_instance(), self)
        # one of the only few scenarios where the command is the same no matter the configuration?
        elif key == 'BackSpace':
            interaction_manager.input_command(['delete_char'], self.graphics, self.get_curr_instance(), self)
        # similar to above
        elif key == 'Return':
            interaction_manager.input_command(['add_new_line'], self.graphics, self.get_curr_instance(), self)

    def user_key_visual(self, key):
        if key in VISUAL_MOVEMENTS:
            motion = VISUAL_MOVEMENTS[key]
            cmd = ['s' + motion[0], 'visual_movement']
            interaction_manager.input_command(cmd, self.graphics, self.get_curr_instance(), self)
            self.command_buffer = ''

    def user_key_ex(self, key):
        if key == 'Return':
            cmd = command_parser.ex_parse(self.command_buffer)
            interaction_manager.input_command(cmd, self.graphics, self.get_curr_instance(), self)
            self.curr_state = 'Default'
            self.command_buffer = ''
        elif key == 'BackSpace':
            self.command_buffer = self.command_buffer[:-1]
        else:
            self.command_buffer += key
    # BEGIN PLUGIN DEFINED ROUTING FUNCITONS HERE
    def init_fuzzy_matching(self):
        if self.curr_state != 'fuzzy_file_selection':
            self.curr_state = 'fuzzy_file_selection'
            self.command_buffer = ''
            self.get_curr_instance().set_visual_anchor(y=2)
            cmd = ['s' + self.command_buffer, 'fuzzy_file_select']
            interaction_manager.input_command(cmd, self.graphics, self.get_curr_instance(), self)

    def user_key_fuzzy_file_select(self, key):
        if key == 'Return':
            self.command_buffer = ''
            self.curr_state = 'Default'
            cmd = ['fuzzy_file_enter']
        elif key == 'BackSpace':
            self.command_buffer = self.command_buffer[:-1]
            cmd = ['s' + self.command_buffer, 'fuzzy_file_select']
        elif key == '<Up>' or key == '<Down>':
            inst = self.get_curr_instance()
            _, vy, _ = inst.get_visual_anchors()
            vy = vy + 1 if key == '<Down>' else vy - 1
            vy = min(21, max(vy, 2))
            inst.set_visual_anchor(y=vy)
            cmd = ['s' + self.command_buffer, 'fuzzy_file_select']
        else:
            self.command_buffer += key
            cmd = ['s' + self.command_buffer, 'fuzzy_file_select']

        interaction_manager.input_command(cmd, self.graphics, self.get_curr_instance(), self)
    # END BEGIN PLUGIN DEFINED ROUTING FUNCITONS HERE



# an attempt to approximate of pygments parser runtime
from pygments import lex
from pygments.lexers import get_lexer_for_filename, get_lexer_by_name
from pygments.token import Token
from color_config import options

import time

def determine_color(t):
    print t
    if t is Token.Name.Class or t is Token.Name.Function:
        return options['function_name_color']
    elif t is Token.Keyword:
        return options['keyword_color']
    elif t is Token.String or t is Token.Literal.String.Interpol:
        return options['string_color']
    elif t is Token.Comment:
        return options['comment_color']
    elif t is Token.Keyword.Namespace:
        return options['namespace_color']
    else:
        return options['text_color']


l = get_lexer_for_filename('test.c')
def parse(s, l):
    ret_list = []
    start = 0
    for token in lex(s, l):
        color = determine_color(token[0])
        ret_list.append((start, token[1], color))
        start += len(token[1])
    print ret_list


# time1 = time.time()
# parse('i', l)
# parse('im', l)
# parse('imp', l)
# parse('impo', l)
# parse('impor', l)
# parse('import', l)
# parse('import test from lol', l)
# parse('def wat(lolblah):', l)
# parse('for line in gg:', l)
# parse('for lin in gg:', l)
# parse('for li in gg:', l)
# parse('for l in gg:', l)
# time2 = time.time()
# print 'function took %0.3f ms' % ((time2-time1)*1000.0)


parse('static int A_is_a(int cur_c)', l)



from pygments import lex
from pygments.lexers import get_lexer_for_filename, get_lexer_by_name
from pygments.token import Token
# TODO: JANKY DUPLICATE IMPORT
from color_config import options

class syntax_parser():
    # init by getting a lexer for file name
    def __init__(self, filename):
        try:
            self.lexer = get_lexer_for_filename(filename)
        except:
            self.lexer = None

    def parse_string(self, s):
        start = 0
        ret_list = []
        # TODO: MAKE THIS NOT SILLY
        if self.lexer == None:
            return ([(0, s, options['text_color'])])
        for token in lex(s, self.lexer):
            color = self.determine_color(token[0])
            ret_list.append((start, token[1], color))
            start += len(token[1])
        return ret_list

    def determine_color(self, t):
        if t is Token.Name.Class or t is Token.Name.Function:
            return options['function_name_color']
        elif t is Token.Keyword:
            return options['keyword_color']
        elif t is Token.String or t is Token.Literal.String.Interpol:
            return options['string_color']
        elif t is Token.Comment:
            return options['comment_color']
        elif t is Token.Keyword.Namespace:
            return options['namespace_color']
        else:
            return options['text_color']



import re, command_list

def goto_line_num(s):
    ind = s.find('gg')
    if ind == 0:
        return ['move_cursor_begin_file']
    else:
        count = 'n' + s[:ind]
        return [count, 'move_cursor_line_num']


def seek_char(s):
    # assumption is that the regex will only return a string of length 2, seems kind of reasonable
    return ['c' + s[1], 'move_cursor_seek_char']


def repeat_default_movement(s):
    n_arg = re.search('[0-9]*', s).group()
    return ['r' + n_arg, command_list.DEFAULT_MOVEMENTS[s[len(n_arg):]][0]]


def delete_text_movement(s):
    return ['s' + command_list.DEFAULT_MOVEMENTS[s[1:]][0], 'delete_text_movement']


def delete_curr_line(s):
    n_arg = re.search('[0-9]*', s).group()
    if bool(n_arg):
        return ['r' + n_arg, 'delete_curr_line']
    else:
        return ['delete_curr_line']


def yank_curr_line(s):
    return ['yank_curr_line']


def quit(s):
    return ['quit']


def write(s):
    return ['write']


DEFAULT_COMMAND_MAP = {
                          re.compile('[0-9]*gg'): goto_line_num,
                          re.compile('f.'): seek_char,
                          re.compile('[0-9]+[h|j|k|l|\{|\}]'): repeat_default_movement,
                          re.compile('d[h|j|k|l|{|}|w|b|e]'): delete_text_movement,
                          re.compile('[0-9]*dd'): delete_curr_line,
                          re.compile('yy'): yank_curr_line
                      }


def default_parse(s):
    for r, func in DEFAULT_COMMAND_MAP.items():
        s_par = r.search(s)
        if bool(s_par):
            return func(s_par.group())
    return ''


VISUAL_COMMAND_MAP = {
                     }

def visual_parse(s):
    for r, func in VISUAL_COMMAND_MAP.items():
        s_par = r.search(s)
        if bool(s_par):
            return func(s_par.group())
    return ''


EX_COMMAND_MAP = {
                     re.compile('q'): quit,
                     re.compile('w'): write
                 }

# TODO:this won't work for actual commands because of the overlap. Figure out a smarter way to do this please
def ex_parse(s):
    print s
    for r, func in EX_COMMAND_MAP.items():
        s_par = r.search(s)
        if bool(s_par):
            return func(s_par.group())
    return ''





import json
from copy import deepcopy
from SyntaxTokens import syntax_parser

# it should be kind of obvious what class is supposed to do.....
class instance():
    def __init__(self, filename):
        self.filename = filename
        self.lines = [line for line in open(filename, 'r')]
        self.parser = syntax_parser.syntax_parser(filename)
        self.line_tokens = [self.parser.parse_string(line) for line in open(filename, 'r')]
        self.cursor_x, self.cursor_y, self.curr_top = 0, 0, 0
        self.visual_x, self.visual_y, self.visual_curr_top = 0, 0, 0
        self.meta_data = json.loads(open('.shimdata', 'r').read())
        self.undo_buffer, self.undo_index = [], -1

    def check_repeat_additions(self, diff, last_state):
        return ((diff[0] == '+' and last_state[0] == '+') and diff[1] == last_state[2]['last_addition'] + 1)

    def check_repeat_deletions(self, diff, last_state):
        return ((diff[0] == '-' and last_state[0] == '-') and diff[1] == last_state[1])

    def check_repeat_modifications(self, diff, last_state):
        return ((diff[0] == 'm' and last_state[0] == 'm') and diff[1] == last_state[1])

    def add_to_undo_buffer(self, diff):
        if self.undo_index == len(self.undo_buffer) - 1:
            try:
                last_state = self.undo_buffer[self.undo_index]
            except IndexError:
                self.undo_buffer.append(diff)
                self.undo_index += 1
                return

            # repeat addtions to the same line
            if self.check_repeat_additions(diff, last_state):
                last_state[2]['count'] += 1
                last_state[2]['last_addition'] += 1
            # repeat deletions from the same line
            elif self.check_repeat_deletions(diff, last_state):
                last_state[2]['lines'].append(diff[2]['lines'][0])
                last_state[2]['line_tokens'].append(diff[2]['line_tokens'][0])
            # repeat modifications to the same line
            elif self.check_repeat_modifications(diff, last_state):
                last_state[2]['new']['line'] = diff[2]['new']['line']
                last_state[2]['new']['line_token'] = diff[2]['new']['line_token']
            else:
                self.undo_buffer.append(diff)
                self.undo_index += 1
                if len(self.undo_buffer) > 100:
                    self.undo_buffer.pop(0)
        else:
            self.undo_buffer = (self.undo_buffer[:self.undo_index] + [diff], [diff])[self.undo_index == -1]
            self.undo_index += 1

    def undo_line_modification(self, diff):
        self.lines[diff[1]] = diff[2]['old']['line'][0]
        self.line_tokens[diff[1]] = diff[2]['old']['line_token'][0]

    def undo_line_addition(self, diff):
        for i in range(diff[2]['count']):
            self.lines.pop(diff[1])
            self.line_tokens.pop(diff[1])

    def undo_line_removal(self, diff):
        for i in range(len(diff[2]['lines'])):
            self.lines.insert(i + diff[1], diff[2]['lines'][i])
            self.line_tokens.insert(i + diff[1], diff[2]['line_tokens'][i])

    def undo_state(self):
        if self.undo_index != -1:
            # diff = self.undo_buffer.pop(-1)
            diff = self.undo_buffer[self.undo_index]
            self.undo_index -= 1
            if diff[0] == 'm':
                self.undo_line_modification(diff)
            elif diff[0] == '+':
                self.undo_line_addition(diff)
            elif diff[0] == '-':
                self.undo_line_removal(diff)

            (x, y, z)  = diff[2]['state']
            self.set_curr_top(z)
            self.set_cursor(x, y)

    def redo_line_modification(self, diff):
        self.lines[diff[1]] = diff[2]['new']['line'][0]
        self.line_tokens[diff[1]] = diff[2]['new']['line_token'][0]

    def redo_line_addition(self, diff):
        for i in range(diff[2]['count']):
            self.lines.insert(diff[1] + i, diff[2]['data']['lines'][i])
            self.line_tokens.insert(diff[1] + i, diff[2]['data']['line_tokens'][i])

    def redo_line_removal(self, diff):
        for i in range(len(diff[2]['lines'])):
            self.lines.pop(diff[1])
            self.line_tokens.pop(diff[1])

    def redo_state(self):
        if self.undo_index < len(self.undo_buffer) - 1:
            self.undo_index += 1
            diff = self.undo_buffer[self.undo_index]
            if diff[0] == 'm':
                self.redo_line_modification(diff)
            if diff[0] == '+':
                self.redo_line_addition(diff)
            if diff[0] == '-':
                self.redo_line_removal(diff)

    def get_line(self, index):
        return self.lines[index]

    # line numbers are 0 indexed
    def get_lines(self):
        return self.lines

    def get_line_tokens(self):
        return self.line_tokens

    def get_cursor(self):
        return self.cursor_x, self.cursor_y

    def get_curr_top(self):
        return self.curr_top

    def get_line_height(self):
        return self.line_height

    def get_line_num(self):
        return len(self.lines)

    def get_filename(self):
        return self.filename

    def get_page_state(self):
        return self.cursor_x, self.cursor_y, self.curr_top

    def get_visual_anchors(self):
        return self.visual_x, self.visual_y, self.visual_curr_top

    def get_meta_data(self):
        return self.meta_data

    def add_line(self, index, line):
        self.add_to_undo_buffer(('+', index,
            {
                'count': 1,
                'data': {
                            'lines': [self.lines[index]],
                            'line_tokens': [self.line_tokens[index]] 
                        },
                'state': self.get_page_state(),
                'last_addition': index 
            }))
        self.lines.insert(index, line)
        self.line_tokens.insert(index, self.parser.parse_string(line))

    def remove_line(self, index):
        self.add_to_undo_buffer(('-', index,
            {
                'lines': [self.lines[index]],
                'line_tokens': [self.line_tokens[index]],
                'state': self.get_page_state(),
            }
        ))
        self.lines.pop(index)
        self.line_tokens.pop(index)

    def set_curr_top(self, num):
        self.curr_top = num

    def set_line_height(self, num):
        self.line_height = num

    def set_line(self, ind, s):
        parsed = self.parser.parse_string(s)
        self.add_to_undo_buffer(('m', ind,
            {
                'old': {
                    'line': [self.lines[ind]],
                    'line_token': [self.line_tokens[ind]] 
                 },
                'new': {
                    'line': [s], 'line_token': [parsed] 
                },
                'state': self.get_page_state()
            }
        ))
        self.lines[ind] = s
        self.line_tokens[ind] = parsed

    def set_visual_anchor(self, x=None, y=None, curr_top=None):
        self.visual_x = x if x is not None else self.cursor_x
        self.visual_y = y if y is not None else self.cursor_y
        self.visual_curr_top = curr_top if curr_top is not None else self.curr_top

    def set_cursor(self, x, y):
        self.cursor_x = max(x, 0)
        if y > self.line_height:
            self.curr_top += (y - self.line_height)
            self.cursor_y = self.line_height
        elif y < 0:
            self.curr_top = max(self.curr_top + y, 0)
            self.cursor_y = 0
        else:
            self.cursor_y = y



from Tkinter import Tk, Canvas, BOTH
from color_config import options
from ttk import Frame
import tkFont

# this is hacky as shit but the math doesn't seem to be working out so far
LINEMAPPING = { 12: 58, 18: 39, 20: 35 }
#LINEMAPPING[font_size]
# Main GUI handler class handles graphics to be displayed to user
class text_canvas(Frame):
    def __init__(self, parent, font_size, input_handler, filename):
        Frame.__init__(self, parent)
        self.parent = parent
        self.text_font = tkFont.Font(family='Monaco', size=font_size, weight='bold')
        self.filename = filename
        self.cheight, self.cwidth, self.line_num_spacing = font_size, self.text_font.measure('c'), 50
        self.line_height = ((self.winfo_screenheight() - self.cheight)/(self.cheight + 2) - 4)
        self.init_UI(input_handler)

    def init_UI(self, input_handler):
        self.parent.title('')
        self.pack(fill=BOTH, expand=1)
        self.init_canvas(input_handler)

    def get_dimensions(self):
        return {
                   'cheight': self.cheight,
                   'cwidth': self.cwidth,
                   'line_num_spacing':self.line_num_spacing,
                   'line_height': self.line_height,
                   'screen_width': self.winfo_screenwidth(),
                   'screen_height': self.winfo_screenheight()
               }

    def init_canvas(self, input_handler):
        self.canvas = Canvas(self, highlightthickness=0, width=self.winfo_screenwidth(), height=self.winfo_screenheight(), bg=options['background_color'])
        self.canvas.pack()
        self.canvas.focus_set()
        self.bind_events(input_handler)

    def clear_all(self):
        self.canvas.delete('all')

    def get_line_height(self):
        return self.line_height

    def get_grid_y(self, y):
        return self.cheight * y + (y * 2)

    def write_line_grid(self, y, line):
        for token in line:
            self.write_text_grid(token[0], y, token[1], token[2])

    # write line of text at given grid co-ordinates
    def write_text_grid(self, x, y, text, color=options['text_color']):
        x_val = self.cwidth * x + self.line_num_spacing
        # 2 pixel spacing between each line
        y_val = self.cheight * y + (y * 2)
        self.canvas.create_text(x_val, y_val, anchor='nw', text=text, font=self.text_font, fill=color)

    def write_status_line(self, text, textcolor=options['status_text_color'], backgroundcolor=options['status_background_color']):
        y = self.line_height + 1
        self.canvas.create_rectangle(0, self.cheight * y + (y * 2), self.winfo_screenwidth(), self.cheight * y + (y * 2) + self.cheight + 4, fill=backgroundcolor, outline=backgroundcolor)
        self.write_text_grid(0, self.line_height + 1, text, textcolor)

    def draw_highlight_grid(self, y, x1, x2, highlightcolor=options['text_highlight_color']):
        y_val = self.cheight * y + (y * 2)
        x1_val = self.cwidth * x1 + self.line_num_spacing
        x2_val = self.cwidth * x2 + self.line_num_spacing
        self.canvas.create_rectangle(x1_val, y_val, x2_val, y_val + self.cheight + 4, fill=highlightcolor, outline=highlightcolor)

    def draw_line_numbers(self, start, highlightcolor=options['line_num_highlight_color'], textcolor=options['line_num_text_color']):
        self.canvas.create_rectangle(0, 0, self.line_num_spacing / 2, self.winfo_screenheight(), fill=highlightcolor, outline=highlightcolor)
        for i in range(self.line_height + 1):
            self.canvas.create_text(0, self.cheight * i + (i * 2), anchor='nw', text=str(start + i), font=self.text_font, fill=textcolor)

    def draw_cursor(self, x, y, highlightcolor=options['cursor_highlight_color'], cursorcolor=options['cursor_color']):
        x_val = self.cwidth * x + self.line_num_spacing
        y_val = self.cheight * y + (y * 2)

        self.canvas.create_rectangle(0, y_val, self.winfo_screenwidth(), y_val + self.cheight + 4, fill=highlightcolor, outline=highlightcolor)
        self.canvas.create_rectangle(x_val, 0, x_val + self.cwidth, self.winfo_screenheight(), fill=highlightcolor, outline=highlightcolor)
        self.canvas.create_rectangle(x_val, y_val, x_val + self.cwidth, y_val + self.cheight + 4, fill=cursorcolor, outline=cursorcolor)

    def draw_rectangle_absolute(self, x1, y1, x2, y2, color):
        self.canvas.create_rectangle(x1, y1, x2, y2, fill=color, outline=color)

    def bind_events(self, input_handler):
        # TODO: this should be cleaned up ideally into a separate handler list
        input_handler.set_GUI_reference(self)
        self.canvas.bind('<Key>', input_handler.key)
        self.canvas.bind_all('<Escape>', input_handler.escape)
        self.canvas.bind_all('<Control-a>', input_handler.control_a)
        self.canvas.bind_all('<Control-b>', input_handler.control_b)
        self.canvas.bind_all('<Control-c>', input_handler.control_c)
        self.canvas.bind_all('<Control-d>', input_handler.control_d)
        self.canvas.bind_all('<Control-e>', input_handler.control_e)
        self.canvas.bind_all('<Control-f>', input_handler.control_f)
        self.canvas.bind_all('<Control-g>', input_handler.control_g)
        self.canvas.bind_all('<Control-h>', input_handler.control_h)
        self.canvas.bind_all('<Control-i>', input_handler.control_i)
        self.canvas.bind_all('<Control-j>', input_handler.control_j)
        self.canvas.bind_all('<Control-k>', input_handler.control_k)
        self.canvas.bind_all('<Control-l>', input_handler.control_l)
        self.canvas.bind_all('<Control-m>', input_handler.control_m)
        self.canvas.bind_all('<Control-n>', input_handler.control_n)
        self.canvas.bind_all('<Control-o>', input_handler.control_o)
        self.canvas.bind_all('<Control-p>', input_handler.control_p)
        self.canvas.bind_all('<Control-q>', input_handler.control_q)
        self.canvas.bind_all('<Control-r>', input_handler.control_r)
        self.canvas.bind_all('<Control-s>', input_handler.control_s)
        self.canvas.bind_all('<Control-t>', input_handler.control_t)
        self.canvas.bind_all('<Control-u>', input_handler.control_u)
        self.canvas.bind_all('<Control-v>', input_handler.control_v)
        self.canvas.bind_all('<Control-w>', input_handler.control_w)
        self.canvas.bind_all('<Control-x>', input_handler.control_x)
        self.canvas.bind_all('<Control-y>', input_handler.control_y)
        self.canvas.bind_all('<Control-z>', input_handler.control_z)
        self.canvas.bind_all("<MouseWheel>", input_handler.mouse_scroll)
        self.canvas.bind_all('<Control-braceright>', input_handler.control_braceright)
        self.canvas.bind_all('<Control-braceleft>', input_handler.control_braceleft)



import cursor_logic


def insert_text_str(s, local_state):
    x, y, curr_top = local_state.get_page_state()
    curr_line = local_state.get_line(curr_top + y)
    local_state.set_line(curr_top + y, curr_line[:x] + s + curr_line[x:])
    local_state.set_cursor(x + len(s), y)


def insert_text_strs(local_state, global_state):
    x, y, curr_top = local_state.get_page_state()
    paste_txt = global_state.get_copy_buffer()

    curr_line = local_state.get_line(curr_top + y)
    local_state.set_line(curr_top + y, curr_line[:x] + paste_txt[0].strip('\n') + curr_line[x:])
    for i in range(len(paste_txt) - 1):
        local_state.add_line(curr_top + y + i + 1, paste_txt[i + 1])


def delete_text_highlight(local_state):
    x, y, curr_top = local_state.get_page_state()
    curr_line = local_state.get_line(curr_top + y)
    local_state.set_line(curr_top + y, curr_line[:x] + curr_line[x + 1:])


def delete_current_line(local_state):
    x, y, curr_top = local_state.get_page_state()
    local_state.remove_line(curr_top + y)
    local_state.set_cursor(0, y)


def insert_new_line_below(local_state):
    x, y, curr_top = local_state.get_page_state()
    curr_line = local_state.get_line(curr_top + y)

    new_line = (' ' * (len(curr_line) - len(curr_line.lstrip()))) + '\n'
    local_state.add_line(y + curr_top + 1, new_line)
    local_state.set_cursor(len(curr_line) - len(curr_line.lstrip()), y + 1)


def insert_new_line_above(local_state):
    x, y, curr_top = local_state.get_page_state()
    curr_line = local_state.get_line(curr_top + y)

    new_line = (' ' * (len(curr_line) - len(curr_line.lstrip()))) + '\n'
    local_state.add_line(y + curr_top, new_line)
    local_state.set_cursor(len(curr_line) - len(curr_line.lstrip()), y)

# TODO: LOOK AT THE MAGIC NUMBERS HERE
def shift_selection_right(local_state):
    px, py, pt = local_state.get_visual_anchors()
    nx, ny, nt = local_state.get_page_state()

    start, end = ((py + pt, ny + nt), (ny + nt, py + pt))[(ny + nt) < (py + pt)]
    for n in range(start, end + 1):
        l = local_state.get_line(n)
        local_state.set_line(n, ' ' * 4 + l)


def shift_selection_left(local_state):
    px, py, pt = local_state.get_visual_anchors()
    nx, ny, nt = local_state.get_page_state()

    start, end = ((py + pt, ny + nt), (ny + nt, py + pt))[(ny + nt) < (py + pt)]
    for n in range(start, end + 1):
        l = local_state.get_line(n)
        local_state.set_line(n, l[:4].strip() + l[4:])


def delete_text_char(local_state):
    x, y, curr_top = local_state.get_page_state()
    curr_line = local_state.get_line(y + curr_top)

    if x > 0:
        local_state.set_line(curr_top + y, curr_line[:x - 1] + curr_line[x:])
        local_state.set_cursor(x - 1, y)
    # do not delete character on first line of page
    elif y > 0 or curr_top > 0:
        if curr_line == '\n':
            local_state.remove_line(curr_top + y)
            local_state.set_cursor(0, y - 1)
            cursor_logic.move_cursor_end_line(local_state)
        else:
            prev_line = local_state.get_line(y + curr_top - 1)
            local_state.remove_line(curr_top + y)
            # slice off new line + last character
            local_state.set_line(curr_top + y - 1, prev_line[:-1] + curr_line)
            local_state.set_cursor(len(prev_line) - 1, y - 1)


def delete_text_range(px, py, pt, nx, ny, nt, local_state):
    # final cursor location should be at fx, fy depending on which comes first
    fx, fy = ((px, py), (nx, ny))[(ny + nt) < (py + pt)]

    if py + pt == ny + nt:
        start, end = ((px, nx), (nx, px))[nx < px]
        curr_line = local_state.get_line(py + pt)
        local_state.set_line(py + pt, curr_line[:start] + curr_line[end:])
        local_state.set_cursor(start, py)
    else:
        start, end = ((py + pt, ny + nt), (ny + nt, py + pt))[(ny + nt) < (py + pt)]
        count = 0
        for n in range(start, end + 1):
            if (n == py + pt) and (px, py) == (fx, fy):
                local_state.set_line(n, local_state.get_line(n)[:px] + '\n')
            elif (n  == py + pt) and (px, py) != (fx, fy):
                local_state.set_line(n, local_state.get_line(n)[px:])
            elif (n == ny + nt) and (nx, ny) == (fx, fy):
                local_state.set_line(n, local_state.get_line(n)[:nx] + '\n')
            elif (n == ny + nt) and (nx, ny) != (fx, fy):
                local_state.set_line(n, local_state.get_line(n)[nx:])
            else:
                count += 1

        for i in range(count):
            local_state.remove_line(start + 1)

        local_state.set_cursor(fx, fy)


def get_text_range(px, py, pt, nx, ny, nt, local_state):
    txt = []
    fx, fy = ((nx, ny), (px, py))[(ny + nt) < (py + pt)]

    if py + pt == ny + nt:
        start, end = ((px, nx), (nx, px))[nx < px]
        curr_line = local_state.get_line(py + pt)
        txt.append(curr_line[start:end + 1])
    else:
        start, end = ((py + pt, ny + nt), (ny + nt, py + pt))[(ny + nt) < (py + pt)]
        count = 0
        for n in range(start, end + 1):
            if (n == py + pt) and (px, py) == (fx, fy):
                txt.append(local_state.get_line(n)[:px])
            elif (n  == py + pt) and (px, py) != (fx, fy):
                txt.append(local_state.get_line(n)[px:])
            elif (n == ny + nt) and (nx, ny) == (fx, fy):
                txt.append(local_state.get_line(n)[:nx])
            elif (n == ny + nt) and (nx, ny) != (fx, fy):
                txt.append(local_state.get_line(n)[nx:])
            else:
                txt.append(local_state.get_line(n))
    return txt



def add_new_line_char(local_state):
    x, y, curr_top = local_state.get_page_state()
    curr_line = local_state.get_line(y + curr_top)

    #  this should be safe I think, no need to perform new line check because python string splicing doesn't toss index exceptions if curr_line[x] != '\n':
    local_state.set_line(y + curr_top, curr_line[:x] + '\n')
    local_state.add_line(y + curr_top + 1, curr_line[x:])

    local_state.set_cursor(0, y + 1)



import os, json

def write_data(dirname, file_data, dir_info):
    paths = []

    for dname, _, filenames in dir_info:
        for filename in filenames:
            if filename not in ['.shimdata']:
                paths.append(os.path.join(dname, filename))

    for path in paths:
        try:
            file_data['fuzzy_file_select'][os.path.relpath(path, dirname)] = path
        except KeyError:
            file_data['fuzzy_file_select'] = { os.path.relpath(path, dirname): path }



options = {
              'background_color':'#002B36',
              'text_color':'#839496',
              'status_text_color':'#002B36',
              'status_background_color':'#657b83',
              'text_highlight_color':'#657b83',
              'line_num_highlight_color':'#073642',
              'line_num_text_color':'#839496',
              'cursor_highlight_color':'#073642',
              'cursor_color':'#7C6B69',
              'function_name_color':'#b58900',
              'keyword_color':'#268bd2',
              'string_color':'#2aa198',
              'comment_color':'#586e75'
          }



def highlight_visual_mode(graphics_state, local_state):
    nx, ny, nt = local_state.get_page_state()
    px, py, pt = local_state.get_visual_anchors()

    if py + pt == ny + nt:
        line = local_state.get_line(py + pt)
        graphics_state.draw_highlight_grid(py, px, nx)
        graphics_state.write_text_grid(0, py, line, '#839496')
    else:
        lp = local_state.get_line(py + pt)
        ln = local_state.get_line(ny + nt)

        if py + pt > ny + nt:
            graphics_state.draw_highlight_grid(py + pt - nt, px, 0)
            graphics_state.write_text_grid(0, py + pt - nt, lp, '#839496')
            graphics_state.draw_highlight_grid(ny, nx, len(ln))
            graphics_state.write_text_grid(0, ny, ln, '#839496')
        else:
            graphics_state.draw_highlight_grid(py + pt - nt, px, len(lp))
            graphics_state.write_text_grid(0, py + pt - nt, lp, '#839496')
            graphics_state.draw_highlight_grid(ny, 0, nx)
            graphics_state.write_text_grid(0, ny, ln, '#839496')

        start, end = ((py + pt, ny + nt), (ny + nt, py + pt))[(ny + nt) < (py + pt)]
        for n in range(start - nt + 1, end - nt):
            l = local_state.get_line(n + nt)
            graphics_state.draw_highlight_grid(n, 0, len(l) - 1)
            graphics_state.write_text_grid(0, n, l, '#839496')



import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


setup(
    name='shim',
    version='0.1.0',
    description="Vim for Python",
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'License :: Other/Proprietary License',
    ),
    keywords='shim vim emacs',
    author=' ',
    author_email=' ',
    url='https://github.com/swong15/shim',
    package_dir = {'': 'src'},
    packages=find_packages('src', exclude=('ez_setup', 'examples', 'tests')),
    include_package_data=True,
    install_requires=(
        "pygments",
    ),
    cmdclass={'test': PyTest},
    tests_require=(
    ),
    entry_points={
        'console_scripts': [
            'shim = shim:main',
        ],
    }
)



def move_cursor_beginning_line(local_state):
    x, y = local_state.get_cursor()
    local_state.set_cursor(0, y)

def move_cursor_end_line(local_state):
    x, y = local_state.get_cursor()
    curr_top = local_state.get_curr_top()
    curr_line = local_state.get_line(y + curr_top)
    local_state.set_cursor(len(curr_line) - 2, y)


def move_cursor_past_end_line(local_state):
    curr_top = local_state.get_curr_top()
    x, y = local_state.get_cursor()
    curr_line = local_state.get_line(y + curr_top)
    local_state.set_cursor(len(curr_line) - 1, y)


def move_cursor_end_file(local_state):
    x, y = local_state.get_cursor()
    curr_top = local_state.get_curr_top()
    local_state.set_curr_top(max(0, local_state.get_line_num() - local_state.get_line_height() - 1))
    local_state.set_cursor(0, min(local_state.get_line_num() -1, local_state.get_line_height()))


def move_cursor_begin_file(local_state):
    local_state.set_curr_top(0)
    x, y = local_state.get_cursor()
    local_state.set_cursor(0, 0)


def move_cursor_left(local_state):
    x, y = local_state.get_cursor()
    x = (0, x - 1)[x - 1 > 0]
    local_state.set_cursor(x, y)


def move_cursor_line_num(n, local_state):
    total = local_state.get_line_num()
    per_page = local_state.get_line_height()
    if n > (total - per_page):
        if total < per_page:
            local_state.set_cursor(0, min(local_state.get_line_num() - 1, n - 1))
        else:
            local_state.set_cursor(0, min(local_state.get_line_num() - 1, per_page - (total - n - 1)))
            local_state.set_curr_top(max(0, total - per_page - 2))
    else:
        local_state.set_curr_top(max(0, min(n - 1, local_state.get_line_num() - 1)))
        local_state.set_cursor(0, 0)


def move_cursor_seek_char(c, local_state):
    curr_top = local_state.get_curr_top()
    x, y = local_state.get_cursor()

    curr_line = local_state.get_line(y + curr_top)
    for offset, char, in enumerate(curr_line[x + 1:]):
        if char == c:
            local_state.set_cursor(x + offset + 1, y)
            break


def move_cursor_right(local_state):
    x, y = local_state.get_cursor()
    curr_top = local_state.get_curr_top()
    lines = local_state.get_lines()
    try:
        nxt_char = lines[curr_top + y][x + 1]
    except:
        nxt_char = '\n'
    x = (x, x + 1)[nxt_char != '\n']
    local_state.set_cursor(x, y)


def move_cursor_up(local_state):
    curr_top = local_state.get_curr_top()
    x, y = local_state.get_cursor()

    try:
        previous_line = local_state.get_line(y + curr_top - 1)
        # last character is a new line
        x = min(x, len(previous_line) - 2)
    except IndexError:
         pass

    local_state.set_cursor(x, y - 1)


def move_cursor_next_paragraph(local_state):
    curr_top = local_state.get_curr_top()
    x, y = local_state.get_cursor()
    accept_all = False

    for offset, line_num in enumerate(range(y + curr_top + 1, local_state.get_line_num() - 1)):
        l = local_state.get_line(line_num).strip()
        if (l == '') and accept_all:
            return local_state.set_cursor(0, y + offset + 1)
        elif l != '':
            accept_all = True

    local_state.set_curr_top(max(0, local_state.get_line_num() - local_state.get_line_height() - 1))
    local_state.set_cursor(0,  min(local_state.get_line_num() - 1, local_state.get_line_height()))


def move_cursor_prev_paragraph(local_state):
    curr_top = local_state.get_curr_top()
    x, y = local_state.get_cursor()
    accept_all = False

    for offset, line_num in enumerate(range(y + curr_top - 1, -1, -1)):
        l = local_state.get_line(line_num).strip()
        if (l == '') and accept_all:
            return local_state.set_cursor(0, y - offset - 1)
        elif l != '':
            accept_all = True

    local_state.set_curr_top(0)
    local_state.set_cursor(0, 0)


def move_cursor_down(local_state):
    curr_top = local_state.get_curr_top()
    x, y = local_state.get_cursor()

    try:
        next_line = local_state.get_line(y + curr_top + 1)
        # last character is a new line
        x = min(x, len(next_line) - 2)
    except IndexError:
         pass
    # if curr line + 1 is < total line numbers then move cursor down
    if (curr_top + y + 1) < local_state.get_line_num():
        local_state.set_cursor(x, y + 1)


def move_cursor_next_word_front(local_state):
    curr_top = local_state.get_curr_top()
    x, y = local_state.get_cursor()
    accept_all = False

    # this is definitely not optimal... there must be a smarter way to do this.
    for index, char in enumerate((local_state.get_line(y + curr_top)[x + 1:])):
        if (accept_all and char != ' ') or (char in ["'", '[', ']', '(', ')', '-', '+', '{', '}']):
            return local_state.set_cursor(x + index + 1, y)
        elif char == ' ':
            accept_all = True

    for offset, line_num in enumerate(range(y + curr_top + 1, local_state.get_line_num())):
        l = local_state.get_line(line_num)
        for index, char in enumerate(l):
            if char != ' ':
                return local_state.set_cursor(index, y + offset + 1)


def move_cursor_move_prev_word_front(local_state):
    curr_top = local_state.get_curr_top()
    x, y = local_state.get_cursor()
    accept_all = False

    # Same as above, I'm pretty sure this can be cleaner.
    curr_str = local_state.get_line(y + curr_top)
    for dx in range(x - 1, -1, -1):
        if (curr_str[dx] != ' ' and curr_str[dx - 1] == ' ') or (curr_str[dx] in ["'", '[', ']', '(', ')', '-', '+', '{', '}']):
            return local_state.set_cursor(dx, y)
        elif (curr_str[dx] != ' ') and dx == 0:
            return local_state.set_cursor(dx, y)

    for dy, line_num in enumerate(range(y + curr_top - 1, -1, -1)):
        curr_str = local_state.get_line(line_num)
        for dx in range(len(curr_str) - 1, 0, -1):
            if (curr_str[dx] != ' ' and curr_str[dx - 1] == ' ') or (curr_str[dx] in ["'", '[', ']', '(', ')', '-', '+', '{', '}']):
                return local_state.set_cursor(dx, y - dy - 1)


def move_cursor_next_word_end(local_state):
    curr_top = local_state.get_curr_top()
    x, y = local_state.get_cursor()
    accept_all = False
    curr_str = local_state.get_line(curr_top + y)

    for dx in range(x + 2, len(curr_str) - 1):
        if (curr_str[dx] == ' ' and curr_str[dx - 1] != ' ') or (curr_str[dx] in ["'", '[', ']', '(', ')', '-', '+', '{', '}']):
            return local_state.set_cursor(dx - 1, y)

    for dy, line_num in enumerate(range(y + curr_top + 1, local_state.get_line_num())):
        curr_str = local_state.get_line(line_num)
        for dx in range(len(curr_str)):
            if (curr_str[dx] == ' ' and curr_str[dx - 1] != ' ') or (curr_str[dx] in ["'", '[', ']', '(', ')', '-', '+', '{', '}']):
                return local_state.set_cursor(dx - 1, y + dy + 1)



from interaction_managers import cursor_logic, text_logic, graphics_logic

# routes keyboard input to appropriate interaction manager to mutate instance state, page is then re-rendered given new state
# events are fed directly from user_input
# interaction manager should not have to parse user input keys directly

def render_default_graphics(graphics_state, local_state, global_state):
    lines = local_state.get_line_tokens()
    x, y, curr_top = local_state.get_page_state()
    buff_line_count = graphics_state.get_line_height()

    graphics_state.draw_cursor(x, y)
    # top is zero indexed and line numbers are one indexed
    graphics_state.draw_line_numbers(curr_top + 1)
    for i in range(buff_line_count + 1):
        # Might have index errors. catch them and move on with life for now
        try:
            graphics_state.write_line_grid(i, lines[curr_top + i])
        except IndexError:
            break
    x, y, t = local_state.get_page_state()
    status_line = 'file: %s  |  mode: %s  |  %d, %d | command buffer: %s' % (local_state.get_filename(), global_state.curr_state, x, y + t, global_state.command_buffer)
    graphics_state.write_status_line(status_line)


def render_page(pre, post, graphics_state, local_state, global_state):
    graphics_state.clear_all()
    for func in pre:
        func()
    render_default_graphics(graphics_state, local_state, global_state)
    for func in post:
        func()


def move_left(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_left(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_right(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_right(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_down(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_down(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_up(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_up(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_beginning_line(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_beginning_line(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_end_line(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_end_line(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_next_word_front(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_next_word_front(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_next_word_end(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_next_word_end(local_state)
    render_page([], [], graphics_state, local_state, global_state)

def move_prev_word_front(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_move_prev_word_front(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_end_file(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_end_file(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_begin_file(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_begin_file(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_next_paragraph(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_next_paragraph(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_prev_paragraph(graphics_state, local_state, global_state):
    cursor_logic.move_cursor_prev_paragraph(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_line_num(n_arg, graphics_state, local_state, global_state):
    cursor_logic.move_cursor_line_num(n_arg, local_state)
    render_page([], [], graphics_state, local_state, global_state)


def move_seek_char(c_arg, graphics_state, local_state, global_state):
    cursor_logic.move_cursor_seek_char(c_arg, local_state)
    render_page([], [], graphics_state, local_state, global_state)


def insert_text(s_arg, graphics_state, local_state, global_state):
    text_logic.insert_text_str(s_arg, local_state)
    render_page([], [], graphics_state, local_state, global_state)


def delete_char(graphics_state, local_state, global_state):
    text_logic.delete_text_char(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def add_new_line(graphics_state, local_state, global_state):
    text_logic.add_new_line_char(local_state)
    render_page([], [], graphics_state, local_state, global_state)

def delete_text_movement(movement, graphics_state, local_state, global_state):
    px, py, pt = local_state.get_page_state()
    COMMAND_MAP[movement](graphics_state, local_state, global_state)
    nx, ny, nt = local_state.get_page_state()

    text_logic.delete_text_range(px, py, pt, nx, ny, nt, local_state)
    render_page([], [], graphics_state, local_state, global_state)


def delete_text_highlight(graphics_state, local_state, global_state):
    if global_state.curr_state == 'Visual':
        px, py, pt = local_state.get_visual_anchors()
        nx, ny, nt = local_state.get_page_state()
        global_state.add_copy_buffer(text_logic.get_text_range(px, py, pt, nx, ny, nt, local_state))
        text_logic.delete_text_range(px, py, pt, nx, ny, nt, local_state)
        # this makes sense to be set here i think
        global_state.curr_state = 'Default'
    else:
        text_logic.delete_text_highlight(local_state)

    render_page([], [], graphics_state, local_state, global_state)


def delete_curr_line(graphics_state, local_state, global_state):
    text_logic.delete_current_line(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def insert_new_line_above(graphics_state, local_state, global_state):
    global_state.curr_state = 'Insert'
    text_logic.insert_new_line_above(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def insert_new_line_below(graphics_state, local_state, global_state):
    global_state.curr_state = 'Insert'
    text_logic.insert_new_line_below(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def insert_end_of_line(graphics_state, local_state, global_state):
    global_state.curr_state = 'Insert'
    cursor_logic.move_cursor_past_end_line(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def mouse_scroll(delta, graphics_state, local_state, global_state):
    x, y, curr_top = local_state.get_page_state()
    if y + int(delta) + curr_top <= local_state.get_line_num() - 2:
        local_state.set_cursor(x, y + int(delta))
        render_page([], [], graphics_state, local_state, global_state)
    else:
        move_end_file(graphics_state, local_state, global_state)


def visual_movement(motion, graphics_state, local_state, global_state):
    COMMAND_MAP[motion](graphics_state, local_state, global_state)
    # some commands break out of visual mode
    if global_state.curr_state == 'Visual':
        render_page([], [lambda: graphics_logic.highlight_visual_mode(graphics_state, local_state)], graphics_state, local_state, global_state)


def paste(graphics_state, local_state, global_state):
    text_logic.insert_text_strs(local_state, global_state)
    render_page([], [], graphics_state, local_state, global_state)


def yank_curr_line(graphics_state, local_state, global_state):
    x, y, curr_top = local_state.get_page_state()
    global_state.add_copy_buffer([local_state.get_line(curr_top + y)])


def shift_selection_right(graphics_state, local_state, global_state):
    text_logic.shift_selection_right(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def shift_selection_left(graphics_state, local_state, global_state):
    text_logic.shift_selection_left(local_state)
    render_page([], [], graphics_state, local_state, global_state)


def quit(graphics_state, local_state, global_state):
    import sys
    sys.exit(1)


def write(graphics_state, local_state, global_state):
    lines = ''.join(local_state.get_lines())
    with open(local_state.get_filename(), 'w') as f:
        f.write(lines)


def undo_command(graphics_state, local_state, global_state):
    local_state.undo_state()
    render_page([], [], graphics_state, local_state, global_state)


def redo_command(graphics_state, local_state, global_state):
    local_state.redo_state()
    render_page([], [], graphics_state, local_state, global_state)


def move_next_instance_buffer(graphics_state, local_state, global_state):
    global_state.go_next_instance()

def move_prev_instance_buffer(graphics_state, local_state, global_state):
    global_state.go_prev_instance()

# BEGIN PLUGIN DEFINED FUNCTIONS HERE
# yanked from: http://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
def calculate_edit_distance(s1, s2):
    oneago = None
    thisrow = range(1, len(s2) + 1) + [0]
    for x in xrange(len(s1)):
        twoago, oneago, thisrow = oneago, thisrow, [0] * len(s2) + [x + 1]
        for y in xrange(len(s2)):
            delcost = oneago[y] + 1
            addcost = thisrow[y - 1] + 1
            subcost = oneago[y - 1] + (s1[x] != s2[y])
            thisrow[y] = min(delcost, addcost, subcost)
    return thisrow[len(s2) - 1]


def sort_files(s, files):
    files = [(k, v, calculate_edit_distance(k, s)) for k, v in files]
    files = sorted(files, key= lambda x: x[2])
    return files

# plugins have access to all state variables in the text editor
def draw_matching_file_names(graphics_state, local_state, global_state):
    dims = graphics_state.get_dimensions()
    graphics_state.draw_rectangle_absolute(0, 0, dims['screen_width'], graphics_state.get_grid_y(21), '#657b83')

    _, vy, _ = local_state.get_visual_anchors()
    graphics_state.draw_rectangle_absolute(0, graphics_state.get_grid_y(vy - 1), dims['screen_width'], graphics_state.get_grid_y(vy), '#dc322f')

    files = sort_files(global_state.command_buffer, local_state.get_meta_data()['fuzzy_file_select'].items())

    graphics_state.write_text_grid(0, 0, global_state.command_buffer, color='#002B36')

    for i in range(20):
        graphics_state.write_text_grid(0, i + 1, files[i][0], color='#002B36')

# router calls this logic
def fuzzy_file_select(s, graphics_state, local_state, global_state):
    post = [lambda:draw_matching_file_names(graphics_state, local_state, global_state)]
    render_page([], post, graphics_state, local_state, global_state)

# router calls this logic
def fuzzy_file_enter(graphics_state, local_state, global_state):
    _, vy, _ = local_state.get_visual_anchors()
    filename = sort_files(global_state.command_buffer, local_state.get_meta_data()['fuzzy_file_select'].items())[vy - 2][0]

    global_state.start_instance(filename)
    global_state.curr_instance += 1
    global_state.set_GUI_reference(graphics_state)
# END PLUGIN DEFINED FUNCTIONS HERE

COMMAND_MAP = {
    'quit': quit,
    'write': write,
    'paste': paste,
    'move_cursor_up': move_up,
    'insert_text': insert_text,
    'delete_char': delete_char,
    'undo_command': undo_command,
    'redo_command': redo_command,
    'mouse_scroll': mouse_scroll,
    'add_new_line': add_new_line,
    'move_cursor_left': move_left,
    'move_cursor_down': move_down,
    'move_cursor_right': move_right,
    'yank_curr_line': yank_curr_line,
    'visual_movement': visual_movement,
    'delete_curr_line': delete_curr_line,
    'move_cursor_end_line': move_end_line,
    'move_cursor_end_file': move_end_file,
    'move_cursor_line_num': move_line_num,
    'move_cursor_seek_char': move_seek_char,
    'insert_end_of_line': insert_end_of_line,
    'move_cursor_begin_file': move_begin_file,
    'shift_selection_left': shift_selection_left,
    'delete_text_movement': delete_text_movement,
    'insert_new_line_above': insert_new_line_above,
    'insert_new_line_below': insert_new_line_below,
    'delete_text_highlight': delete_text_highlight,
    'shift_selection_right': shift_selection_right,
    'move_cursor_next_word_end': move_next_word_end,
    'move_cursor_next_paragraph': move_next_paragraph,
    'move_cursor_prev_paragraph': move_prev_paragraph,
    'move_cursor_beginning_line': move_beginning_line,
    'move_cursor_next_word_front': move_next_word_front,
    'move_cursor_prev_word_front': move_prev_word_front,
    'move_next_instance_buffer': move_next_instance_buffer,
    'move_prev_instance_buffer': move_prev_instance_buffer,
    # BEGIN PLUGIN DEFINED REFERENCES HERE
    'fuzzy_file_select': fuzzy_file_select,
    'fuzzy_file_enter': fuzzy_file_enter,
    # END PLUGIN DEFINED REFERENCES HERE
}


def input_command(command, graphics_state, local_state, global_state):
    if len(command) == 1:
        COMMAND_MAP[command[0]](graphics_state, local_state, global_state)
    else:
        input_command_arg(command, graphics_state, local_state, global_state)


# c denotes character arguments i.e fa maps to find a
# n denotes numerical arguments i.e 123gg maps to jump to line 123
# r denotes repeat arguments i.e 3j means run the 'j' command 3 times
# s denotes character arguments i.e text insert
def input_command_arg(commands, graphics_state, local_state, global_state):
    opt_arg = commands[0][1:]
    in_arg = commands[1]
    if commands[0].startswith('n'):
        COMMAND_MAP[in_arg](int(opt_arg), graphics_state, local_state, global_state)
    elif commands[0].startswith('r'):
        for i in range(int(opt_arg)):
            COMMAND_MAP[in_arg](graphics_state, local_state, global_state)
    elif commands[0].startswith('c'):
        # This should be a single character argument anyway
        COMMAND_MAP[in_arg](opt_arg, graphics_state, local_state, global_state)
    elif commands[0].startswith('s'):
        COMMAND_MAP[in_arg](opt_arg, graphics_state, local_state, global_state)
